
from Graphics import Graphics
from PSMatrix import PSMatrix
from CoordinateSystem import CoordinateSystem
from StringInsert import StringInsert
import Arc
import Fstr 
import math	
import DeviceFont
from Font import Font
import Tex
from BuildChar import BuildChar
from Cmd import *
from Path import Path
import Arrows
import sys

class NoCurrentPoint(Exception):

	def __init__(self):
		pass

	def __str__(self):
		return "\n\n----- No current point!  You probably forgot to begin a path with 'moveto'.\n"
		

# import re

"""
	A Canvas is a simple graphics structure.  The basic datum is the
	array of commands + arguments "cmd".  All commands in PiScript
	are rendered through Canvases.  Coordinate changes are passed up to
	the super-class Graphics; drawing commands are first translated into some
	default coordinates (by means of the current coordinate system)
	before being put into the command array.  

	In any PiSCript program there are three layers of commands.
	The first is coordinate changes, dealt with by Graphics; none of these
	are passed into the output.  In the middle is Canvas, which stores
	the basic command array.  These are meant to be passed to output
	along with begin/endpage and finish by an output class.  This last class
	is the only one that depends on the kind of device serving as output. 

	A Canvas can be placed or embedded in another Canvas.  
	The difference is whether or not it conforms to the current graphics environment.

	StringInserts are strings with metric data.

"""

class Canvas(Graphics):

	""" This list from Cmd.py:

	SETLINEWIDTH = 0 
	SETLINECAP = 1 
	SETLINEJOIN = 2 
	SETDASH = 3 
	SETMITERLIMIT = 4

	SETCOLOR = 5
	SCALELINEWIDTH = 6 
	NEWPATH = 7 
	MOVETO = 8 # gfx
	LINETO = 9 # gfx 

	CURVETO = 10 # gfx 
	FILL = 11
	STROKE = 12 
	CFILL = 13
	CSTROKE = 14

	CLIP = 15
	SETFONT = 16 
	SHOW = 17 # gfx 
	EMBED = 18
	CLOSEPATH = 19 

	INSERT = 20 # gfx?
	IMPORTPS = 21 # gfx
	IMPORTEPS = 22 # gfx
	GSAVE = 23
	GRESTORE = 24

	SHFILL = 25
	COMMENT = 26 """


	"""
			setlinewidth: number
			setlinecap: 0/1/2
			setlinejoin: 0/1/2
			setdash: [pattern] offset
			setmiterlimit: number

			setcolor: r g b
			scalelinewidth: number
			newpath: <nought>
			moveto: x y
			lineto: x y

			curveto: x1 y1 x2 y2 x3 y3
			fill: <nought>
			stroke: <nought>
			cfill: r g b
			cstroke: r g b
			clip: <nought>
			setfont: fontname size
			show: string/Canvas initial_transform_matrix
			embed: <one argument; at the moment this command is not implemented>
			closepath: <nought>

			include: string: PostScript string to include in output
			PSinclude: filename
			EPSinclude: filename
			gsave: <nought>
			grestore: <nought>

			shfill data_source_array
	"""

	# length of command + arguments:
	""" Skip = [
			2, 2, 2, 3, 2,
			4, 2, 1, 3, 3,
			7, 1, 1, 4, 4,
			1, 3, 3, 2, 1,
			2, 3, 3, 1, 1,
			2, 2, 
		] """

	def __init__(self):
		Graphics.__init__(self)
		self.fontTable = DeviceFont.FontTable()
		# self.origin = origin
		self.cmd = []
		# self.m = [1, 0, 0, 1, -origin[0], -origin[1] ] 
		self.m = [1,0,0,1,0,0]
		# pre-multiplies all coordinates
		# The Canvas produced by a dvi file has an arbitrary offset, 
		# which is its origin
		self.comment("canvas init")
		self.setlinewidth(1)
		self.setlinecap(0)
		self.setlinejoin(0)
		self.setdash([], 0)
		self.setmiterlimit(10.0)
		# self.setcolor([0,0,0])
		
		"""
SETLINEWIDTH = 0 
SETLINECAP = 1 
SETLINEJOIN = 2 
SETDASH = 3 
SETMITERLIMIT = 4

SETCOLOR = 5
SCALELINEWIDTH = 6 
NEWPATH = 7 
MOVETO = 8 # gfx
LINETO = 9 # gfx 

CURVETO = 10 # gfx 
FILL = 11
STROKE = 12 
CFILL = 13
CSTROKE = 14

CLIP = 15
SETFONT = 16 
SHOW = 17 # gfx 
EMBED = 18
CLOSEPATH = 19 

INSERT = 20 # gfx?
IMPORTPS = 21 # gfx
IMPORTEPS = 22 # gfx
GSAVE = 23
GRESTORE = 24

SHFILL = 25
	"""
	def clone(self):
		c = Canvas()
		c.comment(" --- clone --- ")
		c.fontTable = self.fontTable
		# c.origin = self.origin
		c.cmd = self.cmd
		c.m = self.m
		c.setlinewidth(self.currentlinewidth())
		c.setlinecap(self.currentlinecap())
		c.setlinejoin(self.currentlinejoin())
		d = self.currentdash()
		c.setdash(d[0], d[1])
		return c
		

	# --- setting up --------------------------------------------------------

	def setmode(self, mode):
		if mode == "graphics":
			self.mode = self.Graphics
		elif mode == "text":
			self.mode = self.Text

	def ctm(self):
		gs = self.cgs()
		return gs.tm

	# both the next two import PS files
	def importPS(self, filename):
		self.cmd.append(IMPORTPS)
		self.cmd.append(filename)
		self.cmd.append(self.ctm())

	def importEPS(self, filename):
		self.cmd.append(IMPORTEPS)
		self.cmd.append(filename)
		self.cmd.append(self.ctm())

	def gsave(self):
		Graphics.gsave(self)
		self.cmd.append(GSAVE)

	def grestore(self):
		Graphics.grestore(self)
		self.cmd.append(GRESTORE)

	# --- internal graphics --------------------------------------------------

	# applies current tm to v
	def image(self, p):
		a = self.ctm()
		p = [ a[0]*p[0] + a[2]*p[1] + a[4], a[1]*p[0] + a[3]*p[1] + a[5] ]
		return p

	# applies current liner componet of tm to v
	def rimage(self, v):
		a = self.ctm()
		v = [ a[0]*v[0] + a[2]*v[1], a[1]*v[0] + a[3]*v[1] ]
		return v

	# --- drawing -------------------------------------------------------------------

	""" 
		NEWPATH = 7 
		MOVETO = 8 
		LINETO = 9 
		CURVETO = 10 
	"""

	def nocurrentpoint(self):
		raise NoCurrentPoint()

	def newpath(self):
		self.setcurrentpoint(None)
		self.cmd.append(NEWPATH)

	# args = ([x, y]) or (x, y)
	def moveto(self, *args):
		if len(args) > 1:
			P = [ args[0], args[1] ]
		else:
			P = args[0]
		self.setlastmove(P)
		self.setcurrentpoint(P)
		Q = self.image(P)
		self.cmd.append(MOVETO)
		self.cmd.append(Q[0])
		self.cmd.append(Q[1])

	# args = ([ x, y]) or (x, y)
	def lineto(self, *args):
		if not self.currentpoint():
			self.nocurrentpoint()
		if len(args) > 1:
			P = [ args[0], args[1] ]
		else:
			P = args[0]
		self.setlastmove(P)
		self.setcurrentpoint(P)
		Q = self.image(P)
		self.cmd.append(LINETO)
		self.cmd.append(Q[0])
		self.cmd.append(Q[1])

	# args = ([ x1, y1, x2, y2, x3, y3]) or (P1, P2, P3)
	def curveto(self, *args):
		if not self.currentpoint():
			self.nocurrentpoint()
		if len(args) == 3:
			P1 = self.image(args[0])
			P2 = self.image(args[1])
			P3 = self.image(args[2])
			self.setcurrentpoint(args[2])
		else:
			P1 = self.image([ args[0], args[1] ])
			P2 = self.image([ args[2], args[3] ])
			P3 = self.image([ args[4], args[5] ])
			self.setcurrentpoint([args[4], args[5]])
		self.cmd.append(CURVETO)
		self.cmd.append(P1[0])
		self.cmd.append(P1[1])
		self.cmd.append(P2[0])
		self.cmd.append(P2[1])
		self.cmd.append(P3[0])
		self.cmd.append(P3[1])

	def closepath(self):
		self.cmd.append(CLOSEPATH)
	
	# --- realizing ----------------------------------------------

	"""
		FILL = 11
		STROKE = 12 
		CLIP = 13
	"""

	# args = (), ([r, g, b]), (g), (r, g, b)
	def fill(self, *args):
		self.setcurrentpoint(None)
		if len(args) == 0:
			self.cmd.append(FILL)
			return
		if len(args) == 1:
			a = args[0]
			if self.isarray(a):
				r = a[0]
				g = a[1]
				b = a[2]
			else: # grey
				r = a
				g = a
				b = a
		else:
			r = args[0]
			g = args[1]
			b = args[2]
		self.cmd.append(CFILL)
		self.cmd.append(r)
		self.cmd.append(g)
		self.cmd.append(b)

	# args = (), ([r, g, b]), (g), (r, g, b)
	def stroke(self, *args):
		self.setcurrentpoint(None)
		if len(args) == 0:
			self.cmd.append(STROKE)
			return
		if len(args) == 1:
			a = args[0]
			if self.isarray(a):
				r = a[0]
				g = a[1]
				b = a[2]
			else: # grey
				r = a
				g = a
				b = a
		else:
			r = args[0]
			g = args[1]
			b = args[2]
		self.cmd.append(CSTROKE)
		self.cmd.append(r)
		self.cmd.append(g)
		self.cmd.append(b)

	def clip(self):
		self.cmd.append(CLIP)


	# --- graphics parameters --------------------------------

	"""
		NOT USED!
		SETLINEWIDTH = 0 
		SETLINECAP = 1 
		SETLINEJOIN = 2 
		SETDASH = 3 
		SETREVERT = 4 # NO!
		SETCOLOR = 5
		SCALELINEWIDTH = 6 
	"""

	# w in true points
	def setlinewidth(self, w):
		Graphics.setlinewidth(self, w)
		self.cmd.append(SETLINEWIDTH)
		self.cmd.append(w)

	# n = 0, 1, 2
	def setlinecap(self, n):
		Graphics.setlinecap(self, n)
		self.cmd.append(SETLINECAP)
		self.cmd.append(n)

	# see manual; default = 10.0
	def setmiterlimit(self, x):
		Graphics.setmiterlimit(self, x)
		self.cmd.append(SETMITERLIMIT)
		self.cmd.append(x)

	# n = 0, 1, 2
	def setlinejoin(self, n):
		Graphics.setlinejoin(self, n)
		self.cmd.append(SETLINEJOIN)
		self.cmd.append(n)

	# [ x0 x1 ... ] off
	def setdash(self, a, off):
		Graphics.setdash(self, a, off)
		self.cmd.append(SETDASH)
		self.cmd.append(a)
		self.cmd.append(off)

	# ([r g b]), (g), (r, g, b)
	def setcolor(self, *args):
		if len(args) == 1:
			a = args[0]
			if self.isarray(a):
				r = a[0]
				g = a[1]
				b = a[2]
			else: # grey
				r = a
				g = a
				b = a
		else:
			r = args[0]
			g = args[1]
			b = args[2]
		Graphics.setcolor(self, [r, g, b])
		self.cmd.append(SETCOLOR)
		self.cmd.append(r)
		self.cmd.append(g)
		self.cmd.append(b)

	# s
	def scalelinewidth(self, s):
		Graphics.scalelinewidth(self, s)
		self.cmd.append(SCALELINEWIDTH)
		self.cmd.append(s)

	# --- miscellaneous --------------------------------------

	def insert(self, s): 
		self.cmd.append(INSERT)
		self.cmd.append(s)

	# ---------------------------------------------------------
	# --- text ------------------------------------------------

	"""
		SETFONT = 14 
		SHOW = 15 
		EMBED = 16
	"""

	def setfont(self, fontName, size):
		self.cmd.append(SETFONT)
		f = self.fontTable.findFont(fontName)
		# WAC: this seems to be normally a Type1 font
		# the fontTable here just acts as general lookup tool
		
		self.setcurrentfont(Font(f, fontName, size))
		self.cmd.append( f.uniqueName() )
		self.cmd.append(size)

	# In here "ValueError: Invalid \x escape"
	# FIXME: need this stupid workaround?
	def updatecharlist(s, f):
		if False:
			i = 0
			while i < len(s):
				c = s[i]
				if c == "\\":
					# convert \xxx to an integer
					a = ord(s[i+1]) - ord("0")
					b = ord(s[i+2]) - ord("0")
					c = ord(s[i+3]) - ord("0")
					n = 64*a+8*b+c
					f.useChar( n )
					i += 4
				else:
					f.useChar( ord(c) )
					i += 1
				# other conversions? \(, \)
		else:
			for c in s:
				f.useChar(c)
	updatecharlist=staticmethod(updatecharlist)

	# puts out an array of character indices (into the current font)
	# and a transformation matrix
	def show(self, s):
		if isinstance(s, StringInsert): 
			#FIXME WAC: kill eventually, we can recover metrics from s and tfm
			s = s.string
		if type(s) == str:
			t = []
			for c in s:
				 t.append(ord(c))
			s = t
		if Graphics.isarray(s): # s now should be an array of integers
			if not self.currentpoint():
				self.moveto(0, 0)
			# writes the string transformed by the Canvas' matrix m
			f = self.currentfont().font
			self.updatecharlist( s, f )
			self.cmd.append(SHOW)
			self.cmd.append(s)
			M = PSMatrix.concat(self.m, self.ctm())
			self.cmd.append(M)	
		else:
			print "\n--- Invalid argument in show! ---\n"



	# the essential difference between place and embed 
	# is that place reverts to default coordinates
	# canvas or canvas, P or canvas, x, y
	def place(self, *args):
		canvas = args[0]
		if len(args) == 1:
			P = (0, 0)
		if len(args) == 2:
			P = args[1]
		elif len(args) == 3:
			P = (args[1], args[2])
		self.fontTable.merge( canvas.fontTable )
		self.gsave()
		gs = self.cgs()
		c = canvas.clone()
		v = gs.transform(P)
		c.m = PSMatrix.concat((1,0,0,1, v[0],v[1]), c.m)
		self.cmd.append(EMBED)
		self.cmd.append(c)
		self.grestore() 
	
	def embed(self, canvas):
		self.fontTable.merge( canvas.fontTable )
		self.gsave()
		tm = self.ctm()
		c = canvas.clone()
		c.m = PSMatrix.concat(tm, c.m) 
		# c.m = PSMatrix.concat([1,0,0,1,tm[4], tm[5]], c.m) 
		# c.m = PSMatrix.concat(c.m, [ tm[0], tm[1], tm[2], tm[3], 0, 0]) 
		self.cmd.append(EMBED)
		self.cmd.append(c)
		self.grestore()

	"""
	# FIXME: WAC x, y?
	def embed(self, *args):
		canvas = args[0]
		if len(args) == 1:
			x = 0; y = 0
		if len(args) == 2:
			(x, y) = args[1]
		elif len(args) == 3:
			x = args[1]; y = args[2]
		self.fontTable.merge( canvas.fontTable )
		self.gsave()
		tm = self.ctm()
		c = canvas.clone()
		c.m = PSMatrix.concat([1,0,0,1,tm[4]+x, tm[5]+y], c.m) 
		c.m = PSMatrix.concat(c.m, [ tm[0], tm[1], tm[2], tm[3], 0, 0])
		self.cmd.append(self.EMBED)
		self.cmd.append(c)
		self.grestore() """

	# ----------------------------------------------------------------------------
	# --- shfill ------------------------------------------------------
	# ----------------------------------------------------------------------------

	# data = a single array, called the data source; designed for PostSCript
	# each item in the array is a coloured `triangle' = array of three `enhanced' vertices [ [x y] [r g b] ]
	# appropriate for Type 4 PS shading
	def shfill(self, data):
		self.cmd.append(SHFILL)
		ds = []
		for t in data:
			T = []
			for p in t:
				q = self.image((p[0][0], p[0][1]))
				T.append([q, p[1]])
			ds.append(T)
		self.cmd.append(ds)

	def comment(self, s):
		self.cmd.append(COMMENT)
		self.cmd.append(s)
		
	# ----------------------------------------------------------------------------
	# --- derivative shapes ------------------------------------------------------
	# ----------------------------------------------------------------------------

	# "moveto" with behaviour like "arc"
	def connectto(self, *args):
		if self.currentpoint(): 
			self.lineto(*args)
		else:
			self.moveto(*args)

	def rmoveto(self, *args):
		P = self.currentpoint()
		if not P:
			self.nocurrentpoint()
		if len(args) > 1:
			V = [ args[0], args[1] ]
		else:
			V = args[0]
		self.setrcurrentpoint(V)
		self.moveto(P[0] + V[0], P[1] + V[1])

	def rlineto(self, *args):
		P = self.currentpoint()
		if not P:
			self.nocurrentpoint()
		if len(args) > 1:
			V = [ args[0], args[1] ]
		else:
			V = args[0]
		self.setrcurrentpoint(V)
		self.lineto(P[0] + V[0], P[1] + V[1])

	def quadto(self, *args):
		P0 = self.currentpoint()
		if not P0:
			self.nocurrentpoint()
		if len(args) == 2:
			P1 = args[0]
			P2 = args[1]
		else:
			P1 = [ args[0], args[1] ]
			P2 = [ args[2], args[3] ]
		# print "current point =", P0
		s = 2.0/3; t = 1.0/3
		self.curveto(t*P0[0]+s*P1[0], t*P0[1]+s*P1[1], s*P1[0]+t*P2[0], s*P1[1]+t*P2[1], P2[0], P2[1])

	def rquadto(self, *args):
		P0 = self.currentpoint()
		if not P0:
			self.nocurrentpoint()
		if len(args) == 2:
			V1 = args[0]
			V3 = args[1]
		else:
			V1 = [ args[0], args[1] ]
			V3 = [ args[2], args[3] ]
		V2 = [ V3[0]-V1[0], V3[1]-V1[1] ]
		s = 2.0/3
		P3 = [ P0[0]+V2[0], p0[1]+V2[1] ]
		P1 = [ P0[0]+s*V1[0], P0[1]+s*V1[1] ]
		P2 = [ P3[0]-s*V2[0], P3[1]-s*V2[1] ]
		self.curveto(P1, P2, P3)

	def rcurveto(self, *args):
		P = self.currentpoint()
		if not P:
			self.nocurrentpoint()
		if len(args) == 3:
			V1 = args[0]
			V2 = args[1]
			V3 = args[2]
		else:
			V1 = [ args[0], args[1] ]
			V2 = [ args[2], args[3] ]
			V3 = [ args[4], args[5] ]
		self.curveto(P[0]+V1[0], P[1]+V1[1], P[0]+V2[0], P[1]+V2[1], P[0]+V3[0], P[1]+V3[1])

	def arc(self, *args):
		Arc.arc(self, *args)

	def arcn(self, *args):
		Arc.arcn(self, *args)
		
	def circle(self, *args):
		Arc.circle(self, *args)

	# args = ([ ll, lr, ur, ul]), (ll, lr, ur, ul), (ll, dx, dy), (llx, lly, dx[0], dx[1], dy[0], dy[1]), 
	def parallelogram(self, *args):
		if len(args) == 1:
			b = args[0]
			self.moveto(b[0])
			self.lineto(b[1])
			self.lineto(b[2])
			self.lineto(b[3])
			self.closepath()
		elif len(args) == 4:
			self.moveto(args[0])
			self.lineto(args[1])
			self.lineto(args[2])
			self.lineto(args[3])
			self.closepath()
		elif len(args) == 3:
			P = args[0]
			U = args[1]
			V = args[2]
			self.moveto(P)
			self.rlineto(U)
			self.rlineto(V)
			self.rlineto(-U[0], -V[0])
		else: # 6
			x = args[0]
			y = args[1]
			dx0 = args[2]
			dy0 = args[3]
			dx1 = args[4]
			dy1 = args[5]
			self.moveto(x, y)
			self.rlineto(dx0, dy0)
			self.rlineto(dx1, dy1)
			self.rlineto(-dx0, -dy0)
		self.closepath()

	# (p), (p[0], ... )
	# does not close up
	def polygon(self, *args):
		if len(args) == 1: # one array of points
			P = args[0]
		else: # several points
			P = args
		self.moveto(P[0])
		for i in range(1, len(P)):
			self.lineto(P[i])

	# (ll, w, h), (x y, w, h)
	# this is a rectangular box
	def box(self, *args):
		if (len(args) == 2):
			w = args[0]
			h = args[1]
			x = 0
			y = 0
		else: # x, y, w, h
			x = args[0]
			y = args[1]
			w = args[2]
			h = args[3]
		self.moveto(x, y)
		self.rlineto(w, 0)
		self.rlineto(0, h)
		self.rlineto(-w, 0)
		self.closepath()

	# ([llx, lly, urx, ury]), ([llx, lly], [urx, ury]), (llx, lly, urx, ury)
	def boundedbox(self, *args):
		if len(args) == 1: # [ llx, lly, urx, ury ]
			llx = args[0][0]
			lly = args[0][1]
			urx = args[0][2]
			ury = args[0][3]
		elif (len(args) == 2):
			llx = args[0][0]
			lly = args[0][1]
			urx = args[1][0]
			ury = args[1][1]
		else: # len = 4: x, y, w, h
			llx = args[0]
			lly = args[1]
			urx = args[2]
			ury = args[3]
		self.moveto(llx, lly)
		self.lineto(urx, lly)
		self.lineto(urx, ury)
		self.lineto(llx, ury)
		self.closepath()

	# --- arrows ---------------------------------------------

	def arrow(self, *args):
		return Arrows.arrow(self, *args)

	def openarrow(self, *args):
		return Arrows.openarrow(self, *args)

	def setarrowdims(self, *args):
		global ps
		Arrows.setarrowdims(self, *args)

	def quadarrow(self, *args):
		global ps
		Arrows.quadarrow(self, *args)

	def texarrow(self, *args):
		global ps
		Arrows.texarrow(self, *args)

	def arcarrow(self, *args):
		global ps
		Arrows.arcarrow(self, *args)

	def arcnarrow(self, *args):
		global ps
		Arrows.arcnarrow(self, *args)

	def ArcArrow(self, c, r, a, b):
		Arrows.ArcArrow(self, c, r, a, b)

	def ArcnArrow(self, c, r, a, b):
		Arrows.ArcnArrow(self, c, r, a, b)

	def arrowheadlength(self):
		return Arrows.arrowheadlength()

	# --- end of arrows -----------------------------------------

	# [-N, -N] -> [N, N] at space ds
	def grid(self, N, ds):
		for i in range(-N, N+1):
			self.moveto(-N*ds, i*ds)
			self.lineto( N*ds, i*ds)
			self.moveto(i*ds,-N*ds)
			self.lineto(i*ds, N*ds)

	def dimensions(self, s):
		cf = self.currentfont()
		fn = cf.name
		size = cf.size
		tfm = cf.metrics()
		width = 0
		height = 0
		depth = 0
		for c in s:
			(w, h, d) = tfm.getCharDims(ord(c))
			width += w
			if h > height:
				height = h
			if d > depth:
				depth = d
		return ( width*size, height*size, depth*size)

	def charpath(self, s):
		cf = self.currentfont()
		f = cf.font
		fn = cf.name
		size = cf.size
		tfm = cf.metrics()

		t = []
		for c in s:
			t.append(ord(c))

		""" self.updatecharlist(t, f) 
		# don't need the character as a character! """

		ev = f.encodingVector()
		t1 = f.type1Font()
		if not t1:
			print "Oops! Can't find font", f
			# substitute Courier?
		tm = t1.fontMatrix()

		tm = [ a*size for a in tm ]
		path = []
		x = 0
		w = 0
		for k in t:
			cs = t1.charStringForChar(k)
			p = BuildChar(cs).path
			# this is a special BuildChar path
			# with commands moveto, lineto, curveto, closepath
			i = 0
			while i < len(p):
				c = p[i]
				path.append(c)
				if c == MOVETO or c == LINETO:
					x = p[i+1]; y = p[i+2]
					(x, y) = PSMatrix.transform(tm, (x+w, y))
					path.append(x); path.append(y)
					i += 3
				elif c == CURVETO:
					x = p[i+1]; y = p[i+2]
					(x, y) = PSMatrix.transform(tm, (x+w, y))
					path.append(x); path.append(y)
					x = p[i+3]; y = p[i+4]
					(x, y) = PSMatrix.transform(tm, (x+w, y))
					path.append(x); path.append(y)
					x = p[i+5]; y = p[i+6]
					(x, y) = PSMatrix.transform(tm, (x+w, y))
					path.append(x); path.append(y)
					i += 7
				else: # c == CLOSEPATH
					i += 1
			w += 1000*tfm.getCharWidth(k)
		return path

	""" def append(self, cmd):
		for c in cmd:
			self.cmd.append(c) """

	# ---------------------------------------------------------

	def __str__(self):
		s = "("
		for j in range(6):
			s += Fstr.cstr(self.m[j]) + ", "
		s += ")\n"
		return s + to_string(self.cmd)


# ============================================================================


