import Type1
from EncodingVector import EncodingVector
import FindResource
import copy
from TexAliasDict import *
from TexFontNameDict import *
from CMRFontDict import cmrfontdict

"""
import logging
"""


class DeviceFont:
	def __init__(self, fontPath, embed = True, extract=True ):
		# FIXME (DM 7/9/09) Do we need the fileName entry any more? The path?
		#self.fileName = fileName # The name of the font as appearing in the map file (e.g. cmr10, ptmr)
		self.fontPath = fontPath # The path to the physical font (e.g. fonts/type1/bluesky/cmr10.pfb
		self.PSName = None       # The PostScript name of the font (e.g. CMR10)
		self.shouldEmbed = embed
		self.shouldExtract = extract
		self.charNameList = set() # List of Postscript character names (/space, /a, etc) of used characters
		self._charsUsed = [ False ] * 256
		self.glyphsUsed = set()

	def useChar( self, c ):
		self._charsUsed[c] = True
	
	def useChars( self, cu ):
		scu = self._charsUsed
		self._charsUsed = [ cu[k] or scu[k] for k in range(256) ]

	def charsUsed( self ):
		cu = self._charsUsed
		return [ k for k in xrange(256) if cu[k] ]

	def encodingVector( self ):
		# StanardEncoding
		return EncodingVector();

	def uniqueName( self ):
		return self.PSName

	def type1Font( self ):
		return None

class PFBFont(DeviceFont):
	def __init__(self, fontPath ):
		DeviceFont.__init__( self, fontPath )
		t1font = FindResource.type1FontForPath( fontPath )
		self.PSName = t1font.fontName()

	def __str__(self):
		return "PFBFont %s: %s" % (self.PSName,self.fontPath)

	def encodingVector( self ):
		font = FindResource.type1FontForPath( self.fontPath )
		return font.encodingVector()

	def type1Font( self ):
		return FindResource.type1FontForPath( self.fontPath )

Base13Names = [ "Times-Roman", "Times-Bold", "Times-Italic", "Times-BoldItalic",
                "Helvetica", "Helvetica-Bold", "Helvetica-Oblique", "Helvetica-BoldOblique",
                "Courier", "Courier-Bold", "Courier-Oblique", "Courier-BoldOblique", "Symbol"]

class Base13Font(DeviceFont):
	def __init__(self, fontName ):
		DeviceFont.__init__( self, None )
		self.PSName = fontName
		self.shouldEmbed = False

	def __str__(self):
		return "Base 13 Font %s" % self.PSName

	def encodingVector( self ):
		# Standard Encoding
		return EncodingVector();
		

Base13Fonts = dict( (k,Base13Font(k)) for k in Base13Names )


# FIXME (DM 07/09/09): Is there any reason not to make DeviceFontWithEncoding a 
# bone fide DeviceFont?????  We could then save on the _charsUsed routines

class DeviceFontWithEncoding:
	def __init__(self, deviceFont, encoding, opts=None ):
		self.deviceFont = deviceFont
		self.encodingFile = encoding
		self.opts = opts
		self._charsUsed = [ False ] *256
	
		# The file name of the encoding is likely not the actual name of the the encoding.
		# We need a field to keep track of the latter.
		self.fullEncodingName = None

	def __str__(self):
		return "FontWithEncoding %s, %s" % (str(self.deviceFont),self.encodingFile)

	def useChar( self, c ):
		self._charsUsed[c] = True

	def useChars( self, cu ):
		scu = self._charsUsed
		self._charsUsed = [ (cu[k] or scu[k]) for k in xrange(256) ]

	def charsUsed( self ):
		cu = self._charsUsed
		return [ k for k in xrange(256) if cu[k] ]

	def encodingVector( self ):
		# Use the explicit encoding, if available
		if self.encodingFile:
			encPath = FindResource.getEncoding( self.encodingFile )
			return EncodingVector.fromFile( encPath )
		# Otherwise, default to the font's
		return self.deviceFont.encodingVector()

	def type1Font( self ):
		return self.deviceFont.type1Font()
		
	def uniqueName( self ):
		if( self.deviceFont == None ):
			return None
		if( self.encodingFile == None):
			return self.deviceFont.PSName
		code = hex(hash(self.encodingFile))[2:]
		return self.deviceFont.PSName+"-"+code

class FontTable:

	def __init__( self ):
		self.deviceFonts={}
		self.encodedFonts={}
		self.mapNameToPSName={}
		self.mapFile = FindResource.getMapFile( "dvipdfm" )
		self.aliasdict = aliasdict
		self.texfontname = texfontname

	def __str__(self):
		s = "WAC:FONT TABLEWAC:\nDevice Fonts:\n"
		for f in self.deviceFonts.values():
			s+= ("  "+str(f)+"\n")
		s+="EncodedFonts:\n"
		for f in self.encodedFonts.values():
			s+= ("  "+str(f)+"\n")
		return s

	"""
	def add( self, texName, deviceFont, encoding, opts=None ):
		font = DeviceFontWithEncoding( deviceFont, encoding, opts=opts )
		psName = deviceFont.PSName
		
		self.deviceFonts[deviceFont.fileName] = deviceFont
		self.encodedFonts[texName] = font
		return font"""

	def get( self, texName ):
		return self.encodedFonts.get( texName, None )

	def getDeviceFont( self, psName ):
		return self.deviceFonts.get( psName, None)

	def addDeviceFont( self, font ):
		fontName = font.PSName
		currentFont = self.deviceFonts.get(fontName, None)
		if currentFont:
			return currentFont
		self.deviceFonts[fontName] = font
		return font

	# --- returns a Type1 font, at least so far --------------------------

	# why aren't the methods in this group static?

	def findFont( self, fontName ):
		if fontName[0] == '/':
			fontName = fontName[1:]
		# print "WAC: Looking for", fontName
		# print "Looking for font", fontName

		if self.aliasdict.has_key(fontName):
			fontName = self.aliasdict[fontName]
			# print "\t(which is an alias for", fontName + ")"
		if self.texfontname.has_key(fontName):
			fontName = self.texfontname[fontName]

		f = self._findLoadedFont( fontName )
		if f:
			# print( "Font %s is loaded." % fontName )
			return f

		f = self._findTeXFontCore( fontName )
		if f:
			# print( "Font %s tex is core." % fontName )
			return f

		f = self._findSystemFontCore( fontName )
		if f:
			# print( "Font %s is system." % fontName )
			return f

		# logging.warning( "Font %s not found.  Using Courier." % fontName )
		print( "Font %s not found.  Using Courier." % fontName )
		return self.findFont( "Courier" )

	def _findLoadedFont( self, fontName ):
		f = self.encodedFonts.get( fontName, None )
		if f:
			return f
		
		f = self.deviceFonts.get( fontName, None )
		if f:
			return f
		
	def findSystemFont( self, fontName ):
		f = self._findLoadedFont( fontName );
		if f:
			return f
		return _findSystemFontCore( texFontName )
				
	def _findSystemFontCore( self, fontName ):
		deviceFont = Base13Fonts.get( fontName, None )
		# print "$$$", deviceFont, fontName
		if deviceFont:
			deviceFont = self.addDeviceFont( deviceFont )
			return deviceFont
	
		texFontName = cmrfontdict.get( fontName, None )
		# print "$$$", texFontName, fontName
		if not texFontName:
			return None
		deviceFontPath = FindResource.getPFB( texFontName )

		# print "$$$", deviceFontPath, fontName
		if not deviceFontPath:
			return None
		deviceFont = self.addDeviceFont( PFBFont( deviceFontPath ) )
		# print "$$$", deviceFont, fontName
		return deviceFont

	def findTeXFont( self, texFontName ):
		f = self._findLoadedFont( texFontName );
		if f:
			return f
		return _findTeXFontCore( texFontName )
				
	def _findTeXFontCore( self, texFontName ):
		"""Look up a device font corresponding to the tex font with name "texFontName".  
		Currently a device font is a Type 1 font together with an encoding vector,
		as listed in tex's map file (dvipdfm.map).  If a font is found, it is added
		to the font table automatically.  None is returned if no font is found."""
		
		# Try the map file to see if there is an entry there.  There should be one.
		mapEntry=self.mapFile.getEntry(texFontName)
		# print "WAC: map entry =", mapEntry
#		failSafeFont = 'cmr12'
		if mapEntry == None:
			return None
#			if texFontName == failSafeFont:
#				raise Exception( "Map file missing basic font cmr12.  Aborting." ) 
#			logging.warning( "Map file missing entry for tex font %s.\nUsing %s instead.", texFontName, failSafeFont )
#			return self.findFont( failSafeFont )

		# We have a map entry.  The name entry is either a base 13 PostScript name or a PFB file name.
		# We need to be able to look up a device font by (generic) PostScript name or by map file name.
		# We'll keep our list of device fonts keyed by PostScript name without the slash, and maintain 
		# a dictionary from map file names to PostScript names.  This assumes that if two font names 
		# in the map file refer to two distinct file names each having the same postscript name, they 
		# are really the same font.  Seems reasonable.

		encoding = mapEntry.encoding
		
		psName = self.mapNameToPSName.get( mapEntry.PSName, None )
		if psName:
			# We have seen this deviceFont before with a different encoding.
			encodedFont = DeviceFontWithEncoding( self.deviceFonts[psName], encoding, opts=mapEntry.opts )
			self.encodedFonts[texFontName] = encodedFont
			return encodedFont

		deviceFont = Base13Fonts.get( mapEntry.PSName, None )
		# print "WAC: base 13?", deviceFont
		if deviceFont:
			deviceFont = self.addDeviceFont( deviceFont )
			return deviceFont

		if not deviceFont:
			# The map entry refers to a file name.  Find it.  Returns the path to the pFB file
			deviceFontPath = FindResource.getPFB( mapEntry.PSName )
			# print "WAC: OK?", deviceFontPath, mapEntry.PSName
			if not deviceFontPath:
				"""
				logging.warning( "Unable to find a device font for\ntex font %s with file name %s.\nUsing %s instead.",dvifont.fontName, fontFileName, failSafeFont ) """
				print ("Unable to find a device font for\ntex font %s with file name %s.\nUsing %s instead.", dvifont.fontName, fontFileName, failSafeFont)
				return self.findFont( failSafeFont )
			deviceFont = PFBFont(deviceFontPath )

		self.mapNameToPSName[mapEntry.PSName] = deviceFont.PSName

		self.deviceFonts[deviceFont.PSName] = deviceFont
		encodedFont = DeviceFontWithEncoding( deviceFont, encoding, opts=mapEntry.opts )
		self.encodedFonts[texFontName] = encodedFont
		return encodedFont

	def merge( self, fontTable ):
		for (k,newFont) in fontTable.encodedFonts.items():
			selfFont = self.encodedFonts.get(k,None)
			if( selfFont ):		
				selfFont.useChars( newFont._charsUsed )		
			else:
				selfFont = newFont
				selfDeviceFont = self.deviceFonts.get( newFont.deviceFont.PSName, None )
				if( selfDeviceFont ):
					selfFont.deviceFont = selfDeviceFont
					selfDeviceFont.useChars( newFont.deviceFont._charsUsed )
				else:
					self.deviceFonts[newFont.deviceFont.PSName] = selfFont.deviceFont
				self.encodedFonts[k] = selfFont
		for (k,v) in fontTable.mapNameToPSName.items():			
			if not self.mapNameToPSName.get(k,None):
				self.mapNameToPSName[k] = v
